// Multiple Render Target, FBO and GLSL example code
//
// A demonstration of using an FBO in your own programs to perform MRT rendering
// with GLSL
// 
// Uses GLee for extension setup
// Note: This isn't meant to indicate good programming style, just act as an example
//

#include "stdafx.h"
#include "glee.h"
#include <gl/glut.h>
#include <cstdlib>

#include "glsl/GLSLProgram.hpp"

GLuint fbo;					// Our handle to the FBO
GLuint depthBuffer;			// Our handle to the depth render buffer
GLuint img, img2;			// Our handle to a textures

const int width = 512;		// The hight of the texture we'll be rendering to
const int height = 512;		// The width of the texture we'll be rendering to

// Used for drawing the 3D cube with our rendered texture on it
GLfloat	xrot = 0;			// X Rotation
GLfloat	yrot = 0;			// Y Rotation
GLfloat xspeed = 0.2f;		// X Rotation Speed
GLfloat yspeed = 0.1f;		// Y Rotation Speed

GLSLProgram * mrtprogram;	// The shader we will be using for MRT rendering output

void init(GLvoid)     
{

	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f, 0.0f, 0.2f, 0.5f);
	glClearDepth(1.0f);					
	glEnable(GL_DEPTH_TEST);			
	glDepthFunc(GL_LEQUAL);				
	glViewport(0,0,800,600);

	// Setup our FBO
	glGenFramebuffersEXT(1, &fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

	// Create the render buffer for depth	
	glGenRenderbuffersEXT(1, &depthBuffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthBuffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, width, height);

	// Now setup the first texture to render to
	glGenTextures(1, &img);
	glBindTexture(GL_TEXTURE_2D, img);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8,  width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// And attach it to the FBO so we can render to it
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, img, 0);

	// after that setup the second texture to render to
	glGenTextures(1, &img2);
	glBindTexture(GL_TEXTURE_2D, img2);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8,  width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// And attach it to the FBO so we can render to it
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, img2, 0);

	// Attach the depth render buffer to the FBO as it's depth attachment
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depthBuffer);

	// Define an array which contains the targets we wish to render to...
	GLenum mrt[] = {GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT};
	// ... then inform OpenGL that we wish to render to these two targets
	glDrawBuffers(2,mrt);

	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
		exit(1);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);	// Unbind the FBO for now

	mrtprogram = new GLSLProgram("mrt.vs","mrt.fs");
}

void ShutDown(GLvoid)
{
	glDeleteFramebuffersEXT(1, &fbo);
	glDeleteRenderbuffersEXT(1, &depthBuffer);
	glDeleteTextures(1,&img);
	glDeleteTextures(1,&img2);

	delete mrtprogram;
}

void reshape(int w,int h)			
{
	glViewport( 0, 0, w, h );
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();					
	if ( h==0 )							
		gluPerspective(80,(float)w,1.0,5000.0);
	else
		gluPerspective(80,(float)w/(float)h,1.0,5000.0);
	glMatrixMode(GL_MODELVIEW);	
	glLoadIdentity();					
}

void keyboard(unsigned char key,int x,int y)  
{
	switch(key)
	{
	case 27:				// When Escape Is Pressed...
		ShutDown();
		exit(0);			// Exit The Program
		break;				
	default:				
		break;
	}
}

void idle(void)
{
	glutPostRedisplay();
}

void display(void)   
{
	// First we bind the FBO so we can render to it
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

	// Save the view port and set it to the size of the texture
	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0,0,width,height);

	// Then render as normal
	// Today's scene is a wonderful multi-coloured spinning cube ;)
	// The destination of the data is controlled by the fragment shader we are using
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
	
	glLoadIdentity();

	glTranslatef(0.0f,0.0f,-2.0f);
	glRotatef(xrot,1.0f,0.0f,0.0f);
	glRotatef(yrot,0.0f,1.0f,0.0f);

	// enable our MRT shader
	mrtprogram->use();
	{
		glBegin(GL_QUADS);
		// Front Face
		glColor4f(0.0f,1.0f,0.0f,1.0f);
		glVertex3f(-0.5f, -0.5,  0.5);
		glVertex3f( 0.5, -0.5,  0.5);
		glVertex3f( 0.5,  0.5,  0.5);
		glVertex3f(-0.5,  0.5,  0.5);
		// Back Face
		glColor4f(1.0f,0.0f,0.0f,1.0f);
		glVertex3f(-0.5, -0.5, -0.5);
		glVertex3f(-0.5,  0.5, -0.5);
		glVertex3f( 0.5,  0.5, -0.5);
		glVertex3f( 0.5, -0.5, -0.5);
		// Top Face
		glColor4f(0.0f,0.0f,1.0f,1.0f);
		glVertex3f(-0.5,  0.5, -0.5);
		glVertex3f(-0.5,  0.5,  0.5);
		glVertex3f( 0.5,  0.5,  0.5);
		glVertex3f( 0.5,  0.5, -0.5);
		// Bottom Face
		glColor4f(0.0f,1.0f,1.0f,1.0f);
		glVertex3f(-0.5, -0.5, -0.5);
		glVertex3f( 0.5, -0.5, -0.5);
		glVertex3f( 0.5, -0.5,  0.5);
		glVertex3f(-0.5, -0.5,  0.5);
		// Right face
		glColor4f(1.0f,1.0f,0.0f,1.0f);
		glVertex3f( 0.5, -0.5, -0.5);
		glVertex3f( 0.5,  0.5, -0.5);
		glVertex3f( 0.5,  0.5,  0.5);
		glVertex3f( 0.5, -0.5,  0.5);
		// Left Face
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glVertex3f(-0.5, -0.5, -0.5);
		glVertex3f(-0.5, -0.5,  0.5);
		glVertex3f(-0.5,  0.5,  0.5);
		glVertex3f(-0.5,  0.5, -0.5);
		glEnd();

	}
	// disable the shader
	mrtprogram->disable();

	// Restore old view port and set rendering back to default frame buffer
	glPopAttrib();
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	// Now we clear the default frame buffer we are going to render to
	glClearColor(0.0f, 0.0f, 0.2f, 0.5f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer

	// Now we get ready to draw the first cube with the first texture attached
	glLoadIdentity();

	// Now bind the first texture to use it
	glBindTexture(GL_TEXTURE_2D, img);
	glEnable(GL_TEXTURE_2D);

	glTranslatef(-1.2f,0.0f,-2.0f);
	glRotatef(-xrot,1.0f,0.0f,0.0f);
	glRotatef(-yrot,0.0f,1.0f,0.0f);

	glColor4f(1.0f,1.0f,1.0f,1.0f);

	// This time it's a textured spinning cube!
	// The texture being the scene we just rendered!
	glBegin(GL_QUADS);
	// Front Face
	glNormal3f( 0.0f, 0.0f, 1.0);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5f, -0.5,  0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0.5, -0.5,  0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0.5,  0.5,  0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5,  0.5,  0.5);
	// Back Face
	glNormal3f( 0.0f, 0.0f,-1.0);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-0.5, -0.5, -0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.5, -0.5, -0.5);
	// Top Face
	glNormal3f( 0.0f, 1.0, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5,  0.5,  0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0.5,  0.5,  0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0.5,  0.5, -0.5);
	// Bottom Face
	glNormal3f( 0.0f,-1.0, 0.0f);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-0.5, -0.5, -0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0.5, -0.5, -0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.5, -0.5,  0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-0.5, -0.5,  0.5);
	// Right face
	glNormal3f( 1.0, 0.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0.5, -0.5, -0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0.5,  0.5,  0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.5, -0.5,  0.5);
	// Left Face
	glNormal3f(-1.0, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5, -0.5, -0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-0.5, -0.5,  0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-0.5,  0.5,  0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5,  0.5, -0.5);
	glEnd();

	// Now we move on to draw the cube again with the 2nd texture
	glLoadIdentity();

	// Now bind the first texture to use it
	glBindTexture(GL_TEXTURE_2D, img2);
	glEnable(GL_TEXTURE_2D);

	glTranslatef(1.2f,0.0f,-2.0f);
	glRotatef(-xrot,1.0f,0.0f,0.0f);
	glRotatef(-yrot,0.0f,1.0f,0.0f);

	glColor4f(1.0f,1.0f,1.0f,1.0f);

	// This time it's a textured spinning cube!
	// The texture being the scene we just rendered!
	glBegin(GL_QUADS);
	// Front Face
	glNormal3f( 0.0f, 0.0f, 1.0);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5f, -0.5,  0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0.5, -0.5,  0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0.5,  0.5,  0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5,  0.5,  0.5);
	// Back Face
	glNormal3f( 0.0f, 0.0f,-1.0);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-0.5, -0.5, -0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.5, -0.5, -0.5);
	// Top Face
	glNormal3f( 0.0f, 1.0, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5,  0.5,  0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0.5,  0.5,  0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0.5,  0.5, -0.5);
	// Bottom Face
	glNormal3f( 0.0f,-1.0, 0.0f);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-0.5, -0.5, -0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0.5, -0.5, -0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.5, -0.5,  0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-0.5, -0.5,  0.5);
	// Right face
	glNormal3f( 1.0, 0.0f, 0.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0.5, -0.5, -0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0.5,  0.5, -0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0.5,  0.5,  0.5);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0.5, -0.5,  0.5);
	// Left Face
	glNormal3f(-1.0, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5, -0.5, -0.5);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-0.5, -0.5,  0.5);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-0.5,  0.5,  0.5);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5,  0.5, -0.5);
	glEnd();

	glDisable(GL_TEXTURE_2D);

	xrot+=xspeed;
	yrot+=yspeed;

	glutSwapBuffers ( );
	// Swap The Buffers To Not Be Left With A Clear Screen
}

int _tmain(int argc, _TCHAR* argv[])
{

	glutInit(&argc, (char**)argv);
	glutInitDisplayMode ( GLUT_RGB | GLUT_DOUBLE );		// Display Mode
	glutInitWindowSize(800,600);
	glutCreateWindow( "FrameBuffer Object Example 2 - Press ESC to exit" );

	// GLee is used to setup all the extensions available
	if(!GLeeInit())
		exit(2);

	init();

	// Setup the various call back functions GLUT requires
	glutDisplayFunc     ( display );  
	glutReshapeFunc     ( reshape );
	glutKeyboardFunc    ( keyboard );
	glutIdleFunc		( idle );
	glutMainLoop        ( );			// Run the main GLUT loop for rendering

	return 0;
}

